// demo_hba.cpp : demonstrate how to use the library of HBA
//
#include <algorithm>
#include <vector>
#include <iostream>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "stdio.h"
#include "stdlib.h"


// the most important header file for RAID controllers
#include "arclib.h"	

// header for SAS HBAs
#include "arclib-sas.h"

#include "ethernet.h"

using namespace std;
// includes the necessary interface header files
#if defined(_WIN32)
	#include "win_comm.h"
	#include "win_pass_ioctl.h"
	#include "win_scsi.h"
    #include "win_ioctl_sas.h"
#elif defined(__linux__) || defined(__FreeBSD__)
	#include "linux_comm.h"
	#include "linux_ioctl.h"
	#if defined(__linux__)
		#include "linux_scsi.h"
		#include "linux_pass_ioctl.h"
        #include "linux_pass_ioctl_sas.h"
	#else
		#include "freebsd_scsi.h"
		#include "freebsd_ioctl_sas.h"
	#endif
#elif defined(__sun__)
	#include "solaris_comm.h"
	#include "solaris_ioctl.h"
	#include "solaris_ioctl_sas.h"
#elif defined(__APPLE__)
	#include "darwin_userclient.h"
    #include "darwin_userclient_sas.h"
#endif


#if (_MSC_VER >= 1600) // vs2010
	#ifdef _WIN64
		#ifdef _DEBUG
			#pragma comment( lib, "..\\..\\bin\\windows\\x86_64\\debug\\arclib.lib" )
		#else
			#pragma comment( lib, "..\\..\\bin\\windows\\x86_64\\release\\arclib.lib" )
		#endif
	#else
		#ifdef _DEBUG
			#pragma comment( lib, "..\\..\\bin\\windows\\i386\\debug\\arclib.lib" )		
		#else
			#pragma comment( lib, "..\\..\\bin\\windows\\i386\\release\\arclib.lib" )	
		#endif
	#endif
#elif (_MSC_VER == 1400) // vs2005
	#ifdef _DEBUG
		#pragma comment( lib, "..\\..\\bin\\win32-vs2005\\debug\\arclib.lib" )
	#else
		#pragma comment( lib, "..\\..\\bin\\win32-vs2005\\release\\arclib.lib" )
	#endif

#endif

#define MAX_CONTROLLER_SUPPORTED	64
CArclibSAS ars[MAX_CONTROLLER_SUPPORTED];




void dumpdata( unsigned char *block, int len)
{
	int ln = (len / 16) + 1;	 // total line#
	unsigned char c;
	int pos = 0;
	 
	printf(" Address = %p, Length = (0x%x)%d\n", block, len, len);
	printf("        0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F      ASCII      \n");
	printf("=======================================================================\n");

	for ( int l = 0; l < ln && len; l++ )
	{
		// printf the line# and the HEX data
		// if a line data length < 16 then append the space to the tail of line to reach 16 chars
		printf("%04X | ", l*0x10);
		for ( pos = 0; pos < 16 && len; pos++, len-- )
		{
			c = block[l*16+pos];    
			printf("%02X ", c);
		}

		if ( pos < 16 )
		{
			for ( int loop = pos; loop < 16; loop++ )
			{
				printf("   ");
			}
		}

		// print ASCII char
		for ( int loop = 0; loop < pos; loop++ )
		{
			c = block[l*16+loop];
			if ( c >= 0x20 && c <= 0x7F )
			{
				printf("%c", c);
			}
			else
			{
				printf(".");
			}
		}
		printf("\n");
	}   
	printf("=======================================================================\n");
}

#if defined(_WIN32)
/************************************************************************************
	Routine Name:	DiscoverySASHBADevice
	Function:		Discover the SAS HBA(s) connected to the PCI, PCI-X, PCI-E bus
	Input:			None.
	printf:			None.
	Return Values:	1. Total HBA(s) found
	Note:			None.
************************************************************************************/
int DiscoverySASHBADevice()
{
	int nTotalCtrl = 0;

	WinIoctlSASInterface *iface = new WinIoctlSASInterface();
	for ( int i = 0; i < 16 ; i++ )
	{
		if ( iface->init(i) )
		{
			ars[nTotalCtrl++].ArcInitSession(iface);
			if ( nTotalCtrl >= MAX_CONTROLLER_SUPPORTED )
			{
				break;
			}

			iface = new WinIoctlSASInterface();                       
		}
	}

	return nTotalCtrl;
}

#endif	// end of defined(_WIN32)


#if defined(__linux__)
int DiscoverySASHBADevice()
{
	int nTotalCtrl = 0;	 

    LinuxPassIoctlSASInterface *iface = new LinuxPassIoctlSASInterface();
    for ( int i = 0; i < 256 ; i++ )
	{
		if ( iface->init(i) )
		{
			ars[nTotalCtrl++].ArcInitSession(iface);
			if ( nTotalCtrl >= MAX_CONTROLLER_SUPPORTED )
			{
				break;
			}

			iface = new LinuxPassIoctlSASInterface();                       
		}
	}
    	
	return nTotalCtrl;
}
#endif	// end of defined(__linux__)


#if defined (__FreeBSD__)
int DiscoverySASHBADevice()
{
	int nTotalCtrl = 0;	 

    FreeBSDIoctlSASInterface *iface = new FreeBSDIoctlSASInterface();
    for ( int i = 0; i < 256 ; i++ )
	{
		if ( iface->init(i) )
		{
			ars[nTotalCtrl++].ArcInitSession(iface);
			if ( nTotalCtrl >= MAX_CONTROLLER_SUPPORTED )
			{
				break;
			}

			iface = new FreeBSDIoctlSASInterface();                       
		}
	}
    	
	return nTotalCtrl;
}
#endif	// end of defined (__FreeBSD__)



#if defined(__sun__)
int DiscoverySASHBADevice()
{
	int nTotalCtrl = 0;	 

    SolarisIoctlSASInterface *iface = new SolarisIoctlSASInterface();
    for ( int i = 0; i < 256 ; i++ )
	{
		if ( iface->init(i) )
		{
			ars[nTotalCtrl++].ArcInitSession(iface);
			if ( nTotalCtrl >= MAX_CONTROLLER_SUPPORTED )
			{
				break;
			}

			iface = new SolarisIoctlSASInterface();                       
		}
	}
    	
	return nTotalCtrl;
}
#endif



#if defined(__APPLE__)
int DiscoverySASHBADevice()
{
	int nTotalCtrl = 0;	

	for ( int i = 0; i < MAX_CONTROLLER_SUPPORTED; i++ )
	{
		DarwinSASUserClientInterface *iface = new DarwinSASUserClientInterface();

		if ( iface->init(i) )
		{
			ars[nTotalCtrl].ArcInitSession(iface);
			nTotalCtrl++;
		}
		else
		{
			delete iface;
			iface = NULL;
		}
	}
	return nTotalCtrl;
}
#endif


char *ReturnHDDType(int hdtype)
{
	switch(hdtype)
	{
	case HD_TYPE_SATA:
		return "SATA";
		break;
	case HD_TYPE_PATA:
		return "PATA";
		break;
	case HD_TYPE_SAS:
		return "SAS";
		break;
	case HD_TYPE_ATAPI:
		return "ATAPI";
		break;
	case HD_TYPE_TAPE:
		return "TAPE";
		break;
	case HD_TYPE_SES:
		return "SES";
		break;
	case HD_TYPE_MEDIA_CHANGER:
		return "Media Changer";
		break;
	default:
		return "Unknown";
		break;
	};

}


char *ReturnDTString(int type)
{
	switch(type)
	{
	case DT_DIRECT_ACCESS_BLOCK:
		return "H.D.D";
		break;
	case DT_SEQ_ACCESS:
		return "Seq Access Device";
		break;
	case DT_PRINTER:
		return "Printer";
		break;
	case DT_PROCESSOR:
		return "Processor";
		break;
	case DT_WRITE_ONCE:
		return "Write Once Device";
		break;
	case DT_CD_DVD:
		return "CD/DVD";
		break;
	case DT_OPTICAL_MEMORY:
		return "Optical Memory";
		break;
	case DT_MEDIA_CHANGER:
		return "Media Changer";
		break;
	case DT_STORAGE_ARRAY_CTRL:
		return "Storage Array Ctrl";
		break;
	case DT_ENCLOSURE:
		return "Enclosure";
		break;
	case DT_EXPANDER:
		return "Expander";
		break;
	case DT_PM:
		return "Port Multiplier";
		break;
	case DT_SES_DEVICE:
		return "SES Device";
		break;
	case DT_LOGICAL_UNIT_NOT_PRESENT_DEVICE:
		return "Logical Unit";
		break;
	default:
		return "Unknown Device";
		break;
	};
}

ARC_STATUS PrintDiskInfo(HD_Info *hdInfo)
{
	ARC_STATUS	status;		
	HD_Config hdconfig;
	PM_Info pmInfo;
	double cap =0.0;
	BYTE szModel[41];
	BYTE szSerialNo[21];
	BYTE szFWVer[9];
	BYTE szProdname[17];
	BYTE szVendorname[16];
	BYTE szRev[16];
	int  encid = 0, enclun = 0;
	int  diskidx = 0;
	int  loop = 0;
	sGUI_ENC_PROPERTY encPro;
	sSENSORS sensors[ARCSAS_MAX_ENCLOSURE_ELEMENT];
	
	if(hdInfo->Size.value == 0)
	{
		status = ARC_NO_SUCH_DEVICE;			
		return status;
	}
		
	memset(&hdconfig, 0, sizeof(HD_Config));
	status = ars[0].SASGetHDConfig(&hdconfig, (BYTE)(hdInfo->Link.Self.DevID & 0xFF), (BYTE)((hdInfo->Link.Self.DevID >> 8) & 0xFF) );
	if(status != ARC_SUCCESS)
	{
		return status;
	}
	
	// HDD information		
	memset(szModel, 0, sizeof(szModel));
    memset(szSerialNo, 0, sizeof(szSerialNo));
    memset(szFWVer, 0, sizeof(szFWVer));
    memcpy(szModel, hdInfo->Model, sizeof(hdInfo->Model));
    memcpy(szSerialNo, hdInfo->SerialNo, sizeof(hdInfo->SerialNo));
    memcpy(szFWVer, hdInfo->FWVersion, sizeof(hdInfo->FWVersion));
	cap = (double)hdInfo->Size.value * (double)0.5 / (double)1024.0 / (double)1024.0;

	printf("The Disk Information\n");
	printf("===============================================================================\n");
	printf("Target ID(ID:LUN)                  : %d:%d\n", hdInfo->Link.Self.DevID & 0xFF, hdInfo->Link.Self.DevID >> 8);
	printf("Model Name                         : %s\n", szModel);
	printf("Serial Number                      : %s\n", szSerialNo);  
	printf("Firmware Rev.                      : %s\n", szFWVer); 	
	if(hdInfo->Link.Parent.DevType == DEVICE_TYPE_EXPANDER)
	{			
		encid = hdInfo->SesEntryIndex & 0xFF;	
		enclun = (hdInfo->SesEntryIndex >> 8) & 0xFF;
		diskidx = hdInfo->SesOverallElementIndex;
		status = ars[0].ArcGetEnclosureProperty( &encPro, encid);
		if ( status != ARC_SUCCESS)
		{				
			return status;
		}

		// enclosure's information
		memset(szProdname, 0, sizeof(szProdname));
		memset(szVendorname, 0, sizeof(szVendorname));
		memset(szRev, 0, sizeof(szRev));	
		memcpy(szProdname, encPro.gepProduct, sizeof(encPro.gepProduct));
		memcpy(szVendorname, encPro.gepVendor, sizeof(encPro.gepVendor));
		memcpy(szRev, encPro.gepRev, sizeof(encPro.gepRev));
		sSENSORS *pPtr = &sensors[encid-ARCSAS_MIN_ENCLOSURE_ID];
		printf("Parent                             : %s %s %s\n", szVendorname, szProdname, szRev);
		printf("Location                           : %s\n", pPtr->disk.snElementStr[diskidx]);
	}
		
	if(hdInfo->Link.Parent.DevType == DEVICE_TYPE_PM)
	{			
		memset(&pmInfo, 0, sizeof(PM_Info));
		status = ars[0].SASGetPMInfo(&pmInfo, hdInfo->Link.Parent.DevID & 0xFF,  (hdInfo->Link.Parent.DevID >> 8) & 0xFF);
		if(status != ARC_SUCCESS)
		{				
			return status;
		}

		printf("Parent                             : Port Multiplier %X:%X\n", pmInfo.VendorId, pmInfo.DeviceId);
		printf("Location                           : SLOT %02d\n", hdInfo->Link.Parent.PhyID[0]);
	}
	printf("SAS Address                        : %02X-%02X-%02X-%02X-%02X-%02X-%02X-%02X\n", 
			                                         hdInfo->Link.Self.SAS_Address[0], 
													 hdInfo->Link.Self.SAS_Address[1],
													 hdInfo->Link.Self.SAS_Address[2],
													 hdInfo->Link.Self.SAS_Address[3],
													 hdInfo->Link.Self.SAS_Address[4],
													 hdInfo->Link.Self.SAS_Address[5],
													 hdInfo->Link.Self.SAS_Address[6],
													 hdInfo->Link.Self.SAS_Address[7]);

	if(hdInfo->DeviceType == DT_DIRECT_ACCESS_BLOCK)
	{		
		printf("Type                               : %s %s\n", ReturnHDDType(hdInfo->HDType), ReturnDTString(hdInfo->DeviceType));
		printf("Capacity                           : %.1fGB(%.lfK)\n", cap, cap*1024.0*1024.0);			
		printf("Write Cache Capability             : %s\n", hdInfo->FeatureSupport & HD_FEATURE_WRITE_CACHE ? "Supported" : "Not Supported");
		printf("S.M.A.R.T Capability               : %s\n", hdInfo->FeatureSupport & HD_FEATURE_SMART ? "Supported" : "Not Supported");
		printf("NCQ Capability                     : %s\n", hdInfo->FeatureSupport & HD_FEATURE_NCQ ? "Supported" : "Not Supported");
		printf("48 Bits LBA Capability             : %s\n", hdInfo->FeatureSupport & HD_FEATURE_48BITS ? "Supported" : "Not Supported");
		printf("Supported Speed                    : %s %s %s\n", hdInfo->FeatureSupport & HD_FEATURE_1_5G ? "1.5Gb/s" : "", 
			                                                   hdInfo->FeatureSupport & HD_FEATURE_3G ? "3.0Gb/s" : "",
															   hdInfo->FeatureSupport & HD_FEATURE_6G ? "6.0Gb/s" : "");			
		printf("Write Cache                        : %s\n", hdconfig.WriteCacheOn ? "Enabled" : "Disabled");
		printf("S.M.A.R.T                          : %s\n", hdconfig.SMARTOn ? "Enabled" : "Disabled");
		const char *szCurrentSpeed[3] = {"1.5Gb/s", "3.0Gb/s", "6.0Gb/s"};
		if(hdconfig.DriveSpeed < SD_SPEED_1_5G || hdconfig.DriveSpeed > SD_SPEED_6G)
		{
			printf("Current Speed                      : %s\n", "Unknown");
		}
		else
		{
			printf("Current Speed                      : %s\n", szCurrentSpeed[hdconfig.DriveSpeed-1]);
		}
		
		if(hdInfo->Rotation_Rate == 0)
		{
			printf("Rotation Rate                      : %s\n", "Rate Not Reported");
		}
		else if(hdInfo->Rotation_Rate == 1)
		{
			printf("Data Set Management(Trim)          : %s\n", hdInfo->FeatureSupport & HD_FEATURE_DATA_SET_MAMAGEMENT ? "Supported" : "Not Supported");
			printf("Rotation Rate                      : %s\n", "Non-Rotating Media(e.g., SSD)");
		}
		else
		{
			printf("Rotation Rate                      : %d\n", hdInfo->Rotation_Rate);
		}
	}
	else
	{
		printf("Type                               : %s\n", ReturnDTString(hdInfo->DeviceType));
	}

	printf("===============================================================================\n");
	return status;
	
}


ARC_STATUS GetHddInfo()
{
	ARC_STATUS      status = ARC_SUCCESS;
	HD_Info         hddInfo;	

	/* Get All HDDs info */
	for(int id = ARCSAS_MIN_HDD_ID; id <= ARCSAS_MAX_HDD_ID; id++)
	{
		memset(&hddInfo, 0, sizeof(HD_Info));
		status = ars[0].SASGetHDInfo(&hddInfo, id);
		if(status == ARC_SUCCESS)
		{		
			PrintDiskInfo(&hddInfo);			
		}
	}	

	return status;
}

ARC_STATUS GetSystemInfo()
{
	ARC_STATUS      status = ARC_SUCCESS;
    Adapter_Info	adInfo;
	Adapter_Config  adconfig;
	const char *PhyRateStr[] = {"Auto", "1.5G", "3.0G", "6.0G", "Unknown", "Unknown", "Unknown", "Unknown"};

	memset(&adInfo, 0, sizeof(Adapter_Info));
	memset(&adconfig, 0, sizeof(Adapter_Config));

	status = ars[0].SASGetHBAInfo(&adInfo);
	if(status != ARC_SUCCESS)
	{		
		printf("status = %x\n", status);
		return status;
	}

	status = ars[0].SASGetHBAConfig(&adconfig);
	if(status != ARC_SUCCESS)
	{		
		printf("status = %x\n", status);
		return status;
	}


	printf("The HBA Information\n");	
	printf("===========================================\n");
	printf("Vendor ID                   : %x\n", adInfo.VenID);
	printf("Device ID                   : %x\n", adInfo.DevID);
	printf("Chip Revision               : %x\n", adInfo.RevisionID);
	printf("# Of Port                   : %d\n", adInfo.PortCount);
	printf("PM(Port Multiplier)         : %s\n", adInfo.MaxPM ? "Supported" : "Not Supported");
	printf("Expander                    : %s\n", adInfo.MaxExpander ? "Supported" : "Not Supported");
	printf("Maximum Supported PM        : %d\n", adInfo.MaxPM);
	printf("Maximum Supported Expander  : %d\n", adInfo.MaxExpander);
	printf("Maximum Supported Device    : %d\n", adInfo.MaxHD);	
	printf("NCQ(Native Command Queuing) : %s\n", adconfig.BIOS_Flag & HBA_FLAG_SATA_NCQ_ENABLE ? "Enabled" : "Disabled");
	printf("S.M.A.R.T Status Polling    : %s\n", adconfig.BIOS_Flag & HBA_FLAG_SMART_STATUS_POLL_ENABLE ? "Enabled" : "Disabled");
	printf("Disk Write Cache            : %s\n", adconfig.BIOS_Flag & HBA_FLAG_DISK_WRITE_CACHE_ENABLE ? "Enabled" : "Disabled");
	printf("Max PHY Link Rate           : %s\n", PhyRateStr[(adconfig.BIOS_Flag >> 8) & 0x07]);
	printf("Max Transfer Length         : %s\n", adconfig.BIOS_Flag & HBA_FLAG_MCXL_4M ? "4M" : "1M");
	printf("BIOS Flags                  : 0x%X\n", adconfig.BIOS_Flag);
	printf("===========================================\n");

	return status;
}


static char *szEventSource[] = 
{
	"MESSAGE",
	"PM SLOT",
	"EXPANDER SLOT",
	"HBA SLOT",
	"PM",
	"EXPANDER",
	"ENCLOSURE",
	"HBA"
};

static char *szEventType[] = 
{
	"NULL",

	/*
	MessageId = 0x1
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_DEVICE_ARRIVAL
	Language = English
	*/
	"Device Added",			

	/*
	MessageId = 0x2
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_DEVICE_REMOVAL
	Language = English
	*/
	"Device Removed",
	
	/*
	MessageId = 0x3
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SATA_ERROR_HANDLE_RETRY
	Language = English
	*/
	"sata error handle retrying request",

	/*
	MessageId = 0x4
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SATA_ERROR_HANDLE_ABORT_REQUST
	Language = English
	*/
	"sata error handle retrying failed, aborting request",

	/*
	MessageId = 0x5
	Severity = Error
	Facility = System
	SymbolicName = ARCSAS_EVENT_SATA_ERROR_HANDLE_PORT_RESET
	Language = English
	*/
	"sata error handle retrying failed, resetting port",

	/*
	MessageId = 0x6
	Severity = Error
	Facility = System
	SymbolicName = ARCSAS_EVENT_SATA_ERROR_HANDLE_SET_DISK_DOWN
	Language = English
	*/
	"sata error handle retrying failed, set disk down",

	/*
	MessageId = 0x7
	Severity = Error
	Facility = System
	SymbolicName = ARCSAS_EVENT_SAS_ERROR_HANDLE_SET_DISK_DOWN
	Language = English
	*/
	"sas error handle retrying failed, set disk down",

	/*
	MessageId = 0x8
	Severity = Error
	Facility = System
	SymbolicName = ARCSAS_EVENT_SAS_ERROR_HANDLE_RESET_COMMAND_SLOT
	Language = English
	*/
	"sas error handle previous task management timeout,reset Port",

	/*
	MessageId = 0x9
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SAS_ERROR_HANDLE_TMF_ABORT_TASK
	Language = English
	*/
	"sas error handle abort task management",

	/*
	MessageId = 0xA
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SAS_ERROR_HANDLE_ABORT_REQUEST
	Language = English
	*/
	"sas error handle abort request",

	/*
	MessageId = 0xB
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SAS_ERROR_HANDLE_TMF_LOGICAL_UNIT_RESET
	Language = English
	*/
	"sas error handle scsi logical unit device reset",

	/*
	MessageId = 0xC
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SAS_ERROR_HANDLE_DEVICE_LINK_RESET
	Language = English
	*/
	"sas error handle scsi device reset",

	/*
	MessageId = 0xD
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SAS_ERROR_HANDLE_DEVICE_HARD_RESET
	Language = English
	*/
	"sas error handle scsi device reset",

	/*
	MessageId = 0xE
	Severity = Error
	Facility = System
	SymbolicName = ARCSAS_EVENT_SAS_ERROR_HANDLE_PORT_RESET
	Language = English
	*/
	"sas error handle sas port reset",

	/*
	MessageId = 0xF
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_PM_ERROR_HANDLE_RETRY_REQUEST
	Language = English
	*/
	"port mutiplier error handle retry request",

	/*
	MessageId = 0x10
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_PM_ERROR_HANDLE_ABORT_REQUEST
	Language = English
	*/
	"port mutiplier error handle retrying failed, aborting request",

	/*
	MessageId = 0x11
	Severity = Error
	Facility = System
	SymbolicName = ARCSAS_EVENT_CORE_PM_ERROR_HANDLE_PORT_RESET
	Language = English
	*/
	"port mutiplier error handle retrying failed, resetting port",

	/*
	MessageId = 0x12
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_CORE_DEVICE_CCB_TIMEOUT
	Language = English
	*/
	"core device ccb timeout",

	/*
	MessageId = 0x13
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_CORE_PM_CCB_TIMEOUT
	Language = English
	*/
	"core port mutiplier ccb timeout",

	/*
	MessageId = 0x14
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SAS_COMMAND_DONE_CORE_HANDLE_TASK_FILE_ERROR
	Language = English
	*/
	"sas command done core handle task file error",

	/*
	MessageId = 0x15
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_PM_HANDLE_DEVICE_UNPLUG
	Language = English
	*/
	"pm handle device unplug but port pPM got null",

	/*
	MessageId = 0x16
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_CORE_INIT_ALLOCATE_CACHED_MEMORY_FAILED
	Language = English
	*/
	"core initialize, allocate core resource error",

	/*
	MessageId = 0x17
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_CORE_INIT_ALLOCATE_UNCACHED_MEMORY_FAILED
	Language = English
	*/
	"core initialize, allocate dma resource error",

	/*
	MessageId = 0x18
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SATA_PM_HOTPLUG_CCB_CALLBACK_READ_PERROR
	Language = English
	*/
	"sata port mutiplier hotplug, hardware register is not correct",

	/*
	MessageId = 0x19
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SATA_PM_HOTPLUG_CCB_CALLBACK_PLUG_IN_STATUS
	Language = English
	*/
	"sata port mutiplier hotplug, plug in status is not correct",
	
	/*
	MessageId = 0x1A
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SATA_PM_HOTPLUG_CCB_CALLBACK_GET_DEVICE_RESOURCE_FAIL
	Language = English
	*/
	"sata port mutiplier hotplug, get null from device pool",

	/*
	MessageId = 0x1B
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_SATA_PM_HOTPLUG_CCB_CALLBACK_PLUG_OUT_STATUS
	Language = English
	*/
	"sata port mutiplier hotplug, plug out status is not correct",

	/*
	MessageId = 0x1C
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_DONE_CCB_SMART_DETECT_ERROR
	Language = English
	*/
	"H.D.D S.M.A.R.T failed",

	/*
	MessageId = 0x1D
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_DONE_CCB_DISABLE_WRITE_CACHE
	Language = English
	*/
	"H.D.D write cache mode disabled",
	
	/*
	MessageId = 0x1E
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_DONE_CCB_ENABLE_WRITE_CACHE
	Language = English
	*/
	"H.D.D write cache mode enabled",

	/*
	MessageId = 0x1F
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_DONE_CCB_DISABLE_SMART
	Language = English
	*/
	"H.D.D S.M.A.R.T disabled",

	/*
	MessageId = 0x20
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_DONE_CCB_ENABLE_SMART
	Language = English
	*/
	"H.D.D S.M.A.R.T enabled",

	/*
	MessageId = 0x21
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_DISK_STATUS_MONITOR_DISABLE
	Language = English
	*/
	"H.D.D S.M.A.R.T status monitor disabled",

	/*
	MessageId = 0x22
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_DISK_STATUS_MONITOR_ENABLE
	Language = English
	*/
	"H.D.D S.M.A.R.T status monitor enabled",

	/*
	MessageId = 0x23
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_DONE_CCB_SMART_POLLING_ERROR
	Language = English
	*/
	"H.D.D S.M.A.R.T polling fail",

	/*
	MessageId = 0x24
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_DONE_CCB_SMART_THRESHOLD_OVER
	Language = English
	*/
	"H.D.D S.M.A.R.T threshold over",

	/*
	MessageId = 0x25
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_CONSTRUCT_AND_SEND_COMMAND_OUT_OF_PRD_ENTRY
	Language = English
	*/
	"Run out of PRD entry, resource not enough",

	/*
	MessageId = 0x26
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_POST_WAITING_LIST_CCB_UNKNOWN_SEND_COMMAND
	Language = English
	*/
	"Unknown send command",

	/*
	MessageId = 0x27
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_DRAIN_DONE_QUEUE_TAG_ERROR
	Language = English
	*/
	"Drain done queue, tag error need service but tag empty",

	/*
	MessageId = 0x28
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_DRAIN_DONE_QUEUE_COMMAND_TABLE_STATUS_ERROR
	Language = English
	*/
	"Drain done queue, pCmdTbl Status Buff Error Info occur",

	/*
	MessageId = 0x29
	Severity = Warning
	Facility = System
	SymbolicName = ARCSAS_EVENT_POST_QUEUE_COMMAND_TABLE_STATUS_ERROR
	Language = English
	*/
	"Post queue, pCmdTbl Status Buff Error Info occur",
	
	/*
	MessageId = 0x2A
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_HBA_DISKS_STATUS_MONITOR_FROM_ON_TO_OFF
	Language = English
	*/
	"H.D.D S.M.A.R.T status monitor stopped",

	/*
	MessageId = 0x2B
	Severity = Informational
	Facility = System
	SymbolicName = ARCSAS_EVENT_HBA_DISKS_STATUS_MONITOR_FROM_OFF_TO_ON
	Language = English
	*/
	"H.D.D S.M.A.R.T status monitor started",
};

ARC_STATUS GetEvent()
{
	ARC_STATUS  status = ARC_SUCCESS;		
	int         nCount = 0;				// total number of event objects we got	
	int         i; 	
	char		timestr[64], devicestr[64], typestr[64], errorstr[64];
	BYTE		parent, child;
	int         nEventFlag = FLAG_ALL;	// the flag pass into ArcGetReqEventPage(...)
	int         nEvtIdx = 0;	

	sEVENT_CODE *ec = NULL;
	sTIMESTAMP  *evtTime = NULL;
	sEVENT_ELEMENT sEventObject;

	
	/* 
		Call ArcGetReqEventPage function to get the required event page from raid subsystem.
		There are 4 event pages reside in raid subsystem and each contains 16 event objects.			
	*/
	status = ars[0].ArcGetReqEventPage(nEventFlag, &nCount);
	if ( status == ARC_SUCCESS )
	{
		status = ars[0].ArcPollEvent(&nEvtIdx);
		if ( nCount != 0 )
		{						
#ifdef _DEBUG
			printf("         1         2         3         4         5         6         7         \n"); 
			printf("1234567890123456789012345678901234567890123456789012345678901234567890123456789\n");
#endif
			printf("Date-Time            Device           Event Type                               \n");
			printf("===============================================================================\n");
			for ( i = 0; i < 128 /*nCount-1*/; i++ )
			{
				memset( timestr, 0, sizeof(timestr));
				memset( devicestr, 0, sizeof(devicestr));
				memset( typestr, 0, sizeof(typestr));				
				memset( errorstr, 0, sizeof(errorstr));

				memset(&sEventObject, 0, sizeof(sEVENT_ELEMENT));				

				/*
					Call ArcGetEventObject function to extract the individual event object from event page
				*/
				if ( ars[0].ArcGetEventObject( (nEvtIdx++) % 128 , &sEventObject ) && 
					 sEventObject.TimeStamp != 0 && 
					 sEventObject.TimeStamp != (-1) &&
					 sEventObject.ErrorCode != (-1) &&
					 sEventObject.EventType != (U8)(-1) )
				{
					memset(timestr, 0, sizeof(timestr));
					evtTime = (sTIMESTAMP *)&sEventObject.TimeStamp;
					if(evtTime->x.tmFlag)
					{						
						/* seconds format */
						struct tm *curtime;
						evtTime->x.tmFlag = 0;
						time_t ltime = (time_t)(evtTime->x.tmTick);
						curtime=localtime(&ltime);						
						sprintf(timestr, "%04d-%02d-%02d %02d:%02d:%02d", 
								curtime->tm_year + 1900, 
								curtime->tm_mon + 1, 
								curtime->tm_mday, 
								curtime->tm_hour, 
								curtime->tm_min,
								curtime->tm_sec);
					}
					else
					{
						sprintf(timestr, "%04d-%02d-%02d %02d:%02d:%02d", 
								evtTime->u.tmYear + 2000, 
								evtTime->u.tmMonth,
								evtTime->u.tmDate,
								evtTime->u.tmHour,
								evtTime->u.tmMinute,
								evtTime->u.tmSecond);
					}
					ec = (sEVENT_CODE *)&sEventObject.ErrorCode;

					switch ( sEventObject.EventType )
					{
					case ARCSAS_EVENT_TYPE_MESSAGE_ONLY:
						ec = (sEVENT_CODE *)&sEventObject.ErrorCode;
						sprintf(devicestr, "Message");
						sprintf(typestr, "%s", szEventType[ec->code] /*, strlen(szEventType[ec->code]) */);
						break;
					case ARCSAS_EVENT_TYPE_PM_SLOT:
						parent = (sEventObject.ParentIndexChildSlot >> 8);
						child = (BYTE)(sEventObject.ParentIndexChildSlot);
						sprintf(devicestr, "PM#%d Slot#%d", parent, child);
						sprintf(typestr, "%s", szEventType[ec->code] /*, strlen(szEventType[ec->code]) */);
						break;
						break;
					case ARCSAS_EVENT_TYPE_EXPANDER_SLOT:
						parent = (sEventObject.ParentIndexChildSlot >> 8);
						child = (BYTE)(sEventObject.ParentIndexChildSlot);
						sprintf(devicestr, "Exp#%d Slot#%d", parent, child);
						sprintf(typestr, "%s", szEventType[ec->code]);
						break;
					case ARCSAS_EVENT_TYPE_HBA_SLOT:
						parent = (sEventObject.ParentIndexChildSlot >> 8);
						child = (BYTE)(sEventObject.ParentIndexChildSlot);
						sprintf(devicestr, "HBA#%d Slot#%d", parent, child);
						sprintf(typestr, "%s", szEventType[ec->code]);
						break;
					case ARCSAS_EVENT_TYPE_PM:
						ec = (sEVENT_CODE *)&sEventObject.ErrorCode;
						sprintf(typestr, "%s", szEventType[ec->code]);
						break;
					case ARCSAS_EVENT_TYPE_EXPANDER:
						parent = (sEventObject.ParentIndexChildSlot >> 8);
						child = (BYTE)(sEventObject.ParentIndexChildSlot);
						sprintf(devicestr, "Exp#%d", parent);
						sprintf(typestr, "%s", szEventType[ec->code]);
						break;
					case ARCSAS_EVENT_TYPE_ENCLOSURE:					
						parent = (sEventObject.ParentIndexChildSlot >> 8);
						child = (BYTE)(sEventObject.ParentIndexChildSlot);
						//sprintf(devicestr, "Exp#%d Enc#%d", parent, sEventObject.DeviceID);
						sprintf(devicestr, "Exp#%d Enc#%d", parent, child);
						sprintf(typestr, "%s", szEventType[ec->code]);
						break;
					case ARCSAS_EVENT_TYPE_HBA:
						ec = (sEVENT_CODE *)&sEventObject.ErrorCode;
						parent = (sEventObject.ParentIndexChildSlot >> 8);
						child = (BYTE)(sEventObject.ParentIndexChildSlot);
						sprintf(devicestr, "HBA#%d", parent);
						sprintf(typestr, "%s", szEventType[ec->code]);
						break;
					default:
						sprintf(typestr, "%s", szEventType[ec->code]);
						break;

					}// end switch					

					printf("%-20s %-16s %-21s %-6s\n", timestr, devicestr, typestr, errorstr);

				}// end if				
			}// end for
			printf("===============================================================================\n");
		}//end if
	}// end if
	
	return status;
}



int Ata_Identify()
{
	ARC_STATUS status = ARC_SUCCESS;
	sATA_PT_CMD atacmd;
	unsigned char data[512];

	memset(&atacmd, 0, sizeof(atacmd));
	memset(data, 0, sizeof(data));

	atacmd.regs.command_status = ATA_IDENTIFY_DEVICE;
	atacmd.command_mode = API_CDB1_ATA_DATA_IN;

	/* demo for scsi-id#0 */
	status = ars[0].ArcAtaPassthroughCmd(&atacmd, 0);
	if(status == ARC_SUCCESS)
	{
		memcpy(data, atacmd.data, 512);
		dumpdata(data, sizeof(data));
	}

	return status;
}


int FindHBA()
{
	int total_ctrl = 0;
	int ctrl = 0;

	/* For SAS HBA Interface */
	ctrl = DiscoverySASHBADevice();
	total_ctrl += ctrl;
	return total_ctrl;
}


ARC_STATUS DemoSASHBA()
{
	int total_ctrl = 0;
	ARC_STATUS  status = ARC_FAILURE;				// see arclib.h for details
	Adapter_Info sysInfo;
	int retval = 1;

	memset(&sysInfo, 0, sizeof(sysInfo));
	
	total_ctrl = FindHBA();
	printf("%d SAS HBA(s) found.\n", total_ctrl);

	if ( total_ctrl )
	{		
		/* warning: use the controller#0 for demonstrating */
		
		/* System Info */
		status = GetSystemInfo();

		/* Get SCSI-ID#0 for demo */
		status = GetHddInfo();

		/* Event */
		status = GetEvent();
	
		/* ATA Passthrough */
		status = Ata_Identify();
	}

	/* IMPORTANT: DO NOT delete the instance of the "Interface" by yourself,
	   library will take care of that */

	return status;
}
